import pycosat
from types import MethodType
import sys

if sys.version.startswith("3.7"):
    from bins.reference37 import PennSAT as reference_PennSAT
elif sys.version.startswith("3.8"):
    from bins.reference38 import PennSAT as reference_PennSAT
elif sys.version.startswith("3.9"):
    from bins.reference39 import PennSAT as reference_PennSAT
else:
    raise RuntimeError("Python version 3.7-3.9 required.")


def test_against_reference(n, cnf, solver_class, test_method=None, **PennSAT_init_kwargs) -> bool:
    """
    Accepts the number of variables `n`, a CNF `cnf`, a solver class (e.g. `PennSAT`),
    and optionally a solver method to test individually (e.g. `test_method=PennSAT.assume`).

    If no solver `test_method` is specified, then this function will run the given solver class
    and PicoSAT on the given CNF. It returns True if the given solver's output matches
    that of PicoSAT (e.g., they are both SAT/both UNSAT), and False otherwise.

    If a solver `test_method` is specified, this function will replace all other solver methods with
    a correct implementation. Essentially, it will only return False if a bug occurs in the
    specified function, rather than anywhere in the code.

    Any extra kwargs will be passed as initialization kwargs to the reference PennSAT
    and `solver_class`.
    """
    ref_soln = pycosat.solve(cnf)
    if test_method is None:
        my_soln = solver_class(n, cnf, **PennSAT_init_kwargs).solve()
    else:
        my_solver = reference_PennSAT(n, cnf, **PennSAT_init_kwargs)
        if test_method.__name__ == "__init__":
            my_solver.__init__ = MethodType(test_method, my_solver)
            my_solver.__init__(n, cnf, **PennSAT_init_kwargs)
        else:
            setattr(my_solver, test_method.__name__, MethodType(test_method, my_solver))
        my_soln = my_solver.solve()
    return (my_soln == ref_soln == "UNSAT") or (my_soln != "UNSAT" and ref_soln != "UNSAT")
